/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003-2004  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "machine.h"
#include "sam.h"
#include "pia.h"
#include "wd2797.h"
#include "logging.h"
#include "types.h"

static Uint8 *addrptr_low;
static Uint8 *addrptr_high;
static Uint32 addrhigh_s;

Uint32 sam_register;

static Uint32 video_base;
static Uint32 video_addr;
static Uint8 *video_ptr;
static Uint32 video_mode;
static Uint32 *video_offsets;
static Uint32 video_increment;
static Uint32 video_ycount;

Uint32 video_offset_list[8][26] = {
	{32,12,0,16,0,16,0,16,0,16,0,16,0,16,0,16,0,16,0,16,0,16,0,16,0,16},
	{16,1,0,0},
	{32,3,0,16,0,16,0,16},
	{48,2,0,0,16,32},
	{32,2,0,16,0,16},
	{16,1,0,0},  /* this is *wrong* for alphanumerics mode */
	{32,1,0,16},
	{42,1,0,16}
};

void sam_init(void) {
	sam_reset();
}

void sam_reset(void) {
	sam_register = 0;
	sam_update_from_register();
	sam_video_vsync();
	video_offsets = video_offset_list[video_mode];
	video_increment = *(video_offsets++);
	video_ycount = *(video_offsets++);
}

Uint32 sam_read_byte(Uint32 addr) {
	if (addr < 0x8000) { return addrptr_low[addr]; }
	if (addr < 0xff00) { return addrptr_high[addr-0x8000]; }
	if (addr < 0xff20) {
		if ((addr & 7) == 0) return pia_data_read(&PIA_0A);
		if ((addr & 7) == 1) return pia_control_read(&PIA_0A);
		if ((addr & 7) == 2) return pia_data_read(&PIA_0B);
		if ((addr & 7) == 3) return pia_control_read(&PIA_0B);
		/* Not yet implemented:
		if ((addr & 7) == 4) return serial_stuff;
		if ((addr & 7) == 5) return serial_stuff;
		if ((addr & 7) == 6) return serial_stuff;
		if ((addr & 7) == 7) return serial_stuff;
		*/
		return 0x7f;
	}
	if (addr < 0xff40) {
		if ((addr & 3) == 0) return pia_data_read(&PIA_1A);
		if ((addr & 3) == 1) return pia_control_read(&PIA_1A);
		if ((addr & 3) == 2) return pia_data_read(&PIA_1B);
		return pia_control_read(&PIA_1B);
	}
	if (addr < 0xff60) {
		/* WD2797 Floppy Disk Controller */
		if ((addr & 15) == 0) return wd2797_status_read();
		if ((addr & 15) == 1) return wd2797_track_register_read();
		if ((addr & 15) == 2) return wd2797_sector_register_read();
		if ((addr & 15) == 3) return wd2797_data_register_read();
		//if (addr & 8) return wd2797_ff48_read();
		return 0x7e;
	}
	if (addr >= 0xffe0) { return rom0[addr-0xc000]; }
	return 0xff;
}

void sam_store_byte(Uint32 addr, Uint32 octet) {
	if (addr < 0x8000) { addrptr_low[addr] = octet; return; }
	if (addr < 0xff00) {
		if (!addrhigh_s) addrptr_high[addr-0x8000] = octet;
		return;
	}
	if (addr < 0xff20) {
		if ((addr & 7) == 0) pia_data_write(&PIA_0A, octet);
		if ((addr & 7) == 1) pia_control_write(&PIA_0A, octet);
		if ((addr & 7) == 2) pia_data_write(&PIA_0B, octet);
		if ((addr & 7) == 3) pia_control_write(&PIA_0B, octet);
		/* Not yet implemented:
		if ((addr & 7) == 4) serial_stuff;
		if ((addr & 7) == 5) serial_stuff;
		if ((addr & 7) == 6) serial_stuff;
		if ((addr & 7) == 7) serial_stuff;
		*/
		return;
	}
	if (addr < 0xff40) {
		if ((addr & 3) == 0) pia_data_write(&PIA_1A, octet);
		if ((addr & 3) == 1) pia_control_write(&PIA_1A, octet);
		if ((addr & 3) == 2) pia_data_write(&PIA_1B, octet);
		if ((addr & 3) == 3) pia_control_write(&PIA_1B, octet);
		return;
	}
	if (addr < 0xff60) {
		/* WD2797 Floppy Disk Controller */
		if ((addr & 15) == 0) wd2797_command_write(octet);
		if ((addr & 15) == 1) wd2797_track_register_write(octet);
		if ((addr & 15) == 2) wd2797_sector_register_write(octet);
		if ((addr & 15) == 3) wd2797_data_register_write(octet);
		if (addr & 8) wd2797_ff48_write(octet);
		return;
	}
	if (addr >= 0xffc0 && addr < 0xffe0) {
		addr -= 0xffc0;
		if (addr & 1)
			sam_register |= 1 << (addr>>1);
		else
			sam_register &= ~(1 << (addr>>1));
		sam_update_from_register();
	}
}

void sam_update_from_register(void) {
	video_mode = sam_register & 0x0007;
	video_base = (sam_register & 0x03f8) << 6;
	if ((addrhigh_s = (~sam_register & 0x8000) >> 15))
		addrptr_high = rom0;
	else
		addrptr_high = ram1;
	addrptr_low = ram0;
	return;
}

void sam_video_pointers(Uint8 **ptrs) {
	*ptrs = video_ptr + *(video_offsets++);
	*(ptrs+1) = video_ptr + *(video_offsets++);
	if (!(--video_ycount)) {
		video_addr += video_increment;
		if (video_addr < 0x8000) {
			video_ptr = addrptr_low + video_addr;
		} else {
			video_ptr = addrptr_high + video_addr - 0x8000;
		}
		video_offsets = video_offset_list[video_mode];
		video_increment = *(video_offsets++);
		video_ycount = *(video_offsets++);
	}
}

void sam_video_vsync(void) {
	video_addr = video_base;
	video_offsets = video_offset_list[video_mode];
	video_increment = *(video_offsets++);
	video_ycount = *(video_offsets++);
	if (video_addr < 0x8000) {
		video_ptr = addrptr_low + video_addr;
	} else {
		video_ptr = addrptr_high + video_addr - 0x8000;
	}
}
